/******************************************************************************
 * (c) Copyright 2000, Devin Butterfield, Robert A. Buaas
 *
 * ixj_isa.c
 *
 * ISA Bus Specific Device Driver Support for FreeBSD. 
 *
 *
 * 		This file provides the isa bus specific probe, attach and
 *		detach functions for the Quicknet Internet PhoneJACK and
 *		LineJACK telephony cards.
 *
 *		This program is free software; you can redistribute it and/or
 *		modify it under the terms of the GNU General Public License
 *		as published by the Free Software Foundation; either version
 *		2 of the License, or (at your option) any later version.
 *
 * Authors:	Devin Butterfield, <dbutter@wireless.net>
 *		Robert A. Buaas, <buaas@wireless.net>
 *
 * Notes:
 * 	-	Support for loadable kernel modules for ISAPNP cards is
 *		currently broken in FreeBSD <= 4.1-RELEASE. This is because
 * 		the "unknown" ISAPNP driver gobbles the resources for devices
 * 		that the kernel does already have a driver for at boot time,
 * 		later preventing loadable modules from getting their needed
 *  		resources. --DB 10/15/2000
 *		
 *	- 	Must add the correct PNP ID's for the original PhoneJACK and
 *		the LineJACK card.
 *	-	The code for the original PhoneJACK and the LineJACK is in
 *		place but has not been tested. Only the PhoneJACK lite has
 *		been tested so far. --DB 10/25/2000
 * 
 * Fixes:
 *
 * $Id: ixj_isa.c,v 1.7 2001/03/03 12:18:34 devinb Exp $
 *
 *****************************************************************************/
 
#include <sys/types.h>			/* Defines for signed/unsigned	*/
#include <sys/param.h>                  /* priorities, min(), etc.	*/
#include <sys/conf.h>			/* cdevsw stuff			*/
#include <sys/kernel.h>                 /* provides ticks, sysinit, etc.*/
#include <sys/module.h>			/* kernel module support	*/
#include <sys/systm.h>			/* for sprintf, etc.		*/
#include <sys/malloc.h>			/* memory allocation		*/
#include <sys/poll.h>			/* poll event definitions	*/
#include <sys/select.h>			/* select routines		*/
#include <sys/uio.h>                    /* moving data from/to kernel	*/
#include <sys/bus.h>                    /* bus definitions/routines	*/
#include <sys/ioccom.h>			/* ioctl macros/definitions	*/
#include <sys/fcntl.h>			/* definitions for file modes	*/
#include <machine/clock.h>		/* For DELAY()			*/
#include <machine/bus.h>		/* for inb(), outb, etc.	*/
#include <machine/resource.h>           /* resource type definitions	*/
#include <machine/ipl.h>		/* for cpl intr mask/ipending	*/
#include <sys/signalvar.h>		/* for async io / signals	*/
#include <sys/filio.h>			/* FIOASYNC, FIONBIO, defs	*/
#include <sys/rman.h>                   /* resource manager routines	*/
#include <isa/isavar.h>			/* more ISA stuff...		*/
#include "ixj.h"			/* ixj structures etc.		*/
#include "ixjvar_bsd.h"                     /* softc, vars, prototypes, etc.*/

#define DSP_RID		0               /* DSP resource id		*/
#define XILINX_RID 	1               /* XILINX resource id		*/

					/* Function prototypes.		*/
static int	ixj_isa_probe	__P((device_t dev));
static int	ixj_isa_attach	__P((device_t dev));
static int	ixj_isa_detach	__P((device_t dev));

                                        /* Possible Vendor ID's		*/
                                        /* add ID's for all cards here	*/
static struct isa_pnp_id ixj_ids[] = {
	{ 0x10018946,	"QTI0110 Internet PhoneJack"      }, /* XXX: must set this to correct id */
	{ 0x10038946,	"QTI0310 Internet LineJack"       }, /* QTI0300 */	
	{ 0x10048946,	"QTI0410 Internet PhoneJack Lite" }, /* QTI0410	*/	
	{ 0,		NULL }
};
					/* inline function		*/
extern __inline__ void LED_SetState(int state, int board)
{
	if (ixj[board].cardtype == QTI_LINEJACK) {
		ixj[board].pld_scrw.bits.led1 = state & 0x1 ? 1 : 0;
		ixj[board].pld_scrw.bits.led2 = state & 0x2 ? 1 : 0;
		ixj[board].pld_scrw.bits.led3 = state & 0x4 ? 1 : 0;
		ixj[board].pld_scrw.bits.led4 = state & 0x8 ? 1 : 0;

		outb_p(ixj[board].pld_scrw.byte, ixj[board].XILINXbase);
	}
}

static int
ixj_isa_probe(device_t dev)             /* isapnp probe routine		*/
{
	int error = 0;
					/* get value from pnp probe	*/
 	error = ISA_PNP_PROBE(device_get_parent(dev), dev, ixj_ids);

                                       	/* if the card had a pnp id	*/
					/* that we don't know...	*/
	if (error == ENXIO)
		return ENXIO;
                                        /* if we had some other error	*/

	if (!(error == 0 || error == ENOENT))
		return (error);
                                        /* if we have the resources	*/
					/* then we're good to go...	*/
	if(bus_get_resource_start(dev, SYS_RES_IOPORT, 0) != 0)
		return (0);

	return (ENXIO);			/* default to fail  		*/
}

static int
ixj_isa_attach(device_t dev)        	/* attach device		*/
{

        struct ixj_softc *sc;
	int unit = device_get_unit(dev);
	int error;
	int dsp_rid;
	int xilinx_rid;
	u_int32_t logical_id;

	sc = (struct ixj_softc *) device_get_softc(dev);

                                        /* initialize softc/ixj		*/
        bzero(sc, sizeof(struct ixj_softc));
	bzero(&ixj[unit], sizeof(IXJ));

	if(unit > (IXJMAX - 1)) {	/* check that we've not exceeded*/
					/* max number of cards...	*/
		device_printf(dev, "More than %d cards not supported!\n", IXJMAX);
		return ENXIO;
	}
	
        logical_id = isa_get_logicalid(dev);

	switch(logical_id) {
	case 0x10018946:		/* XXX: must get correct id for PJACK*/
		ixj[unit].cardtype = QTI_PHONEJACK;
		break;
	case 0x10038946:               	/* Internet LineJACK		*/
		ixj[unit].cardtype = QTI_LINEJACK;
		break;
	case 0x10048946:                /* Internet PhoneJack Lite	*/
		ixj[unit].cardtype = QTI_PHONEJACK_LITE;
		if(ixjdebug > 0)
			device_printf(dev, "cardtype %d detected.\n", 
						ixj[unit].cardtype);
		break;
	}
	
	dsp_rid = 0;	               	/* set rid then alloc for DSP	*/
	sc->dsp_ioport = bus_alloc_resource(dev, SYS_RES_IOPORT, &dsp_rid,
					0ul, ~0ul, 16, RF_ACTIVE);

					/* quit if resource alloc failed*/
	if(!(sc->dsp_ioport)) {
		device_printf(dev, "No I/O space for DSP?\n");
		error = ENXIO;
		goto bad;
	}

                                        /* get DSP port start addr	*/
	sc->DSPbase = rman_get_start(sc->dsp_ioport);
	ixj[unit].DSPbase = sc->DSPbase;

	if(ixjdebug > 0)
		device_printf(dev, "DSPbase at port 0x%x\n", 
					ixj[unit].DSPbase);

					/* reassign rid in case it was	*/
					/* changed by bus_alloc.	*/
	sc->dsp_rid = dsp_rid;

					/* if not old PJACK then there	*/
					/* should be a XILINX PLD.	*/
	if(ixj[unit].cardtype != QTI_PHONEJACK) {
		xilinx_rid = 1;		/* set rid then alloc for XILINX*/
		sc->xilinx_ioport = bus_alloc_resource(dev, SYS_RES_IOPORT,
			       	&xilinx_rid, 0ul, ~0ul, 4, RF_ACTIVE);

					/* quit if resource alloc failed*/
		if(!(sc->xilinx_ioport)) {
			device_printf(dev, "No I/O space for XILINX?\n");
			error = ENXIO;
			goto bad;
		}
					/* get XILINX port start addr	*/
		sc->XILINXbase = rman_get_start(sc->xilinx_ioport);
		ixj[unit].XILINXbase = sc->XILINXbase;
		
		if(ixjdebug > 0)	
			device_printf(dev, "XILINXbase at port 0x%x\n", 
						ixj[unit].XILINXbase);

                                        /* reassign rid in case it was	*/
					/* changed by bus_alloc.	*/
		sc->xilinx_rid = xilinx_rid;
	}	

	error = ixj_attach(unit, sc);	/* do bus independent attach	*/
	if(error) { 			/* quit on failure.		*/
		printf("received %d from ixj_attach!!\n", error);
		goto bad;
	}	

	if(unit == 0)			/* if this is the first attach,	*/
		ixj_add_timer();	/* start main timeout loop.	*/	
	
	return (0);
bad:
	if(sc->dsp_ioport) {		/* release any alloc'ed res	*/
		bus_release_resource(dev, SYS_RES_IOPORT, sc->dsp_rid,
				sc->dsp_ioport);
	}
	if(sc->xilinx_ioport) {		/* release any alloc'ed res	*/
		bus_release_resource(dev, SYS_RES_IOPORT, sc->xilinx_rid,
				sc->xilinx_ioport);
	}
	return ENXIO;
}

static int
ixj_isa_detach (device_t dev)		/* detach device		*/
{
	struct ixj_softc *sc;
        int unit = device_get_unit(dev);

	sc = (struct ixj_softc *) device_get_softc(dev);
	
					/* don't detach if dev is open!	*/
	if (ixj[unit].device_state & IXJ_STATE_OPEN)
		return (EBUSY);

	if (unit == 0)			/* if this is the first detach	*/
		untimeout(ixj_timeout, NULL, ch); /* clear timeout loop	*/ 

	if (ixj[unit].cardtype == QTI_LINEJACK) {
       		     		        /* Turn off DAA Frame Sync 	*/
		ixj[unit].pld_scrw.bits.daafsyncen = 0;
		outb(ixj[unit].XILINXbase, ixj[unit].pld_scrw.byte);
		ixj[unit].pld_slicw.bits.rly1 = 0;
		ixj[unit].pld_slicw.bits.rly2 = 0;
		ixj[unit].pld_slicw.bits.rly3 = 0;
		outb(ixj[unit].XILINXbase + 0x01, ixj[unit].pld_slicw.byte);
		LED_SetState(0x0, unit);
	}
	                               	/* check if resources were	*/
					/* originally allocated...	*/
					/* then release them if so.	*/
	if(ixjdebug > 0)
		device_printf(dev, "Releasing resources...\n");

	if (ixj[unit].read_buffer) {
		if(ixjdebug > 0)
			device_printf(dev, "freeing read buffer...\n");
		free(ixj[unit].read_buffer, M_DEVBUF);
	}
	
	if (ixj[unit].write_buffer) {
		if(ixjdebug > 0)
			device_printf(dev, "freeing write buffer...\n");
		free(ixj[unit].write_buffer, M_DEVBUF);
	}
					/* release io resources		*/
	if(sc->dsp_ioport) {
		if(ixjdebug > 0)
			device_printf(dev, "releasing ioport resources for DSP\n"); 
		bus_release_resource(dev, SYS_RES_IOPORT,
				sc->dsp_rid, sc->dsp_ioport);
	} else {
		device_printf(dev, "***No resources to release for DSP!\n"); 
	}
					/* if not old PJACK then there	*/
					/* should be a XILINX PLD	*/
	if(ixj[unit].cardtype != QTI_PHONEJACK) {
		if(sc->xilinx_ioport) {
			if(ixjdebug > 0)
				device_printf(dev, "releasing ioport resources for XILINX\n");
			bus_release_resource(dev, SYS_RES_IOPORT,
				sc->xilinx_rid, sc->xilinx_ioport);
		} else {
			device_printf(dev, "***No resources to release for XILINX!\n"); 
		}
	}
	
	destroy_dev(sc->dev);		/* remove device entry		*/
	
	return (0);
}

static device_method_t ixj_isa_methods[] = {
	/* device interface	*/
	DEVMETHOD(device_probe,		ixj_isa_probe),
	DEVMETHOD(device_attach,	ixj_isa_attach),
	DEVMETHOD(device_detach,	ixj_isa_detach),
	{ 0, 0 }
};

static driver_t ixj_isa_driver = {	/* the basic device interface	*/
	"ixj",
	ixj_isa_methods,
	sizeof(struct ixj_softc)
};

static	devclass_t	ixj_isa_devclass;

DRIVER_MODULE(ixj, isa, ixj_isa_driver, ixj_isa_devclass, 0, 0);

