/******************************************************************************  * (c) Copyright 2001, Roger Hardiman
 *
 * ixj_pci.c
 *
 * PCI Bus Specific Device Driver Support for FreeBSD. 
 *
 *
 * 		This file provides the PCI bus specific probe, attach and
 *		detach functions for the Quicknet Internet PhoneJACK
 *
 *		This program is free software; you can redistribute it and/or
 *		modify it under the terms of the GNU General Public License
 *		as published by the Free Software Foundation; either version
 *		2 of the License, or (at your option) any later version.
 *
 * Authors:	Roger Hardiman, <roger@freebsd.org>
 *		Based on ixj_isa.c written by
 *		Devin Butterfield, <dbutter@wireless.net>
 *		Robert A. Buaas, <buaas@wireless.net>
 *
 * Notes:
 * 
 * Fixes:
 *
 * $Id: ixj_pci.c,v 1.1 2001/03/26 14:40:22 rogerh Exp $
 *
 *****************************************************************************/
 
#include <sys/types.h>			/* Defines for signed/unsigned	*/
#include <sys/param.h>                  /* priorities, min(), etc.	*/
#include <sys/conf.h>			/* cdevsw stuff			*/
#include <sys/kernel.h>                 /* provides ticks, sysinit, etc.*/
#include <sys/module.h>			/* kernel module support	*/
#include <sys/systm.h>			/* for sprintf, etc.		*/
#include <sys/malloc.h>			/* memory allocation		*/
#include <sys/poll.h>			/* poll event definitions	*/
#include <sys/select.h>			/* select routines		*/
#include <sys/uio.h>                    /* moving data from/to kernel	*/
#include <sys/bus.h>                    /* bus definitions/routines	*/
#include <sys/ioccom.h>			/* ioctl macros/definitions	*/
#include <sys/fcntl.h>			/* definitions for file modes	*/
#include <machine/clock.h>		/* For DELAY()			*/
#include <machine/bus.h>		/* for inb(), outb, etc.	*/
#include <machine/resource.h>           /* resource type definitions	*/
#include <machine/ipl.h>		/* for cpl intr mask/ipending	*/
#include <sys/signalvar.h>		/* for async io / signals	*/
#include <sys/filio.h>			/* FIOASYNC, FIONBIO, defs	*/
#include <sys/rman.h>                   /* resource manager routines	*/
#include "ixj.h"			/* ixj structures etc.		*/
#include "ixjvar_bsd.h"                     /* softc, vars, prototypes, etc.*/

#include <sys/bus.h>
#include <machine/bus_memio.h>  /* used by bus space */
#include <machine/bus.h>        /* used by bus space and newbus */
 
#include <sys/rman.h>           /* used by newbus */
#include <machine/resource.h>   /* used by newbus */

#include <pci/pcivar.h>
#include <pci/pcireg.h>


#define QUICKNET_VENDOR_ID 0x15e2
#define PCI_DEVICE_ID_PHONEJACK 0x0500

#define DSP_RID		0               /* DSP resource id		*/
#define XILINX_RID 	1               /* XILINX resource id		*/

					/* Function prototypes.		*/
static int	ixj_pci_probe	__P((device_t dev));
static int	ixj_pci_attach	__P((device_t dev));
static int	ixj_pci_detach	__P((device_t dev));


static int
ixj_pci_probe(device_t dev)             /* pci probe routine		*/
{
	if (pci_get_vendor(dev) != QUICKNET_VENDOR_ID)
		return (ENXIO);

	if (pci_get_device(dev) == PCI_DEVICE_ID_PHONEJACK) {
		device_set_desc(dev, "Quicknet PhoneJACK PCI");
		return 0;
	}

	return (ENXIO);
}

static int
ixj_pci_attach(device_t dev)        	/* attach device		*/
{

        struct ixj_softc *sc;
	int unit = device_get_unit(dev);
	int error;
	int dsp_rid;

	sc = (struct ixj_softc *) device_get_softc(dev);

                                        /* initialize softc/ixj		*/
        bzero(sc, sizeof(struct ixj_softc));
	bzero(&ixj[unit], sizeof(IXJ));

	if(unit > (IXJMAX - 1)) {	/* check that we've not exceeded*/
					/* max number of cards...	*/
		device_printf(dev, "More than %d cards not supported!\n", IXJMAX);
		return ENXIO;
	}
	
	ixj[unit].cardtype = QTI_PHONEJACK_PCI;
	
	dsp_rid = PCIR_MAPS;		/* set rid then alloc for DSP	*/
	sc->dsp_ioport = bus_alloc_resource(dev, SYS_RES_IOPORT, &dsp_rid,
					0ul, ~0ul, 16, RF_ACTIVE);

					/* quit if resource alloc failed*/
	if(!(sc->dsp_ioport)) {
		device_printf(dev, "No I/O space for DSP?\n");
		error = ENXIO;
		goto bad;
	}

                                        /* get DSP port start addr	*/
	sc->DSPbase = rman_get_start(sc->dsp_ioport);
	ixj[unit].DSPbase = sc->DSPbase;

	sc->XILINXbase       = ixj[unit].DSPbase + 0x10;
	ixj[unit].XILINXbase = ixj[unit].DSPbase + 0x10;

	if(ixjdebug > 0)
		device_printf(dev, "DSPbase at port 0x%x\n", 
					ixj[unit].DSPbase);

					/* reassign rid in case it was	*/
					/* changed by bus_alloc.	*/
	sc->dsp_rid = dsp_rid;

	error = ixj_attach(unit, sc);	/* do bus independent attach	*/
	if(error) { 			/* quit on failure.		*/
		printf("received %d from ixj_attach!!\n", error);
		goto bad;
	}	

	if(unit == 0)			/* if this is the first attach,	*/
		ixj_add_timer();	/* start main timeout loop.	*/	
	
	return (0);
bad:
	if(sc->dsp_ioport) {		/* release any alloc'ed res	*/
		bus_release_resource(dev, SYS_RES_IOPORT, sc->dsp_rid,
				sc->dsp_ioport);
	}
	return ENXIO;
}

static int
ixj_pci_detach (device_t dev)		/* detach device		*/
{
	struct ixj_softc *sc;
        int unit = device_get_unit(dev);

	sc = (struct ixj_softc *) device_get_softc(dev);
	
					/* don't detach if dev is open!	*/
	if (ixj[unit].device_state & IXJ_STATE_OPEN)
		return (EBUSY);

	if (unit == 0)			/* if this is the first detach	*/
		untimeout(ixj_timeout, NULL, ch); /* clear timeout loop	*/ 

	                               	/* check if resources were	*/
					/* originally allocated...	*/
					/* then release them if so.	*/
	if(ixjdebug > 0)
		device_printf(dev, "Releasing resources...\n");

	if (ixj[unit].read_buffer) {
		if(ixjdebug > 0)
			device_printf(dev, "freeing read buffer...\n");
		free(ixj[unit].read_buffer, M_DEVBUF);
	}
	
	if (ixj[unit].write_buffer) {
		if(ixjdebug > 0)
			device_printf(dev, "freeing write buffer...\n");
		free(ixj[unit].write_buffer, M_DEVBUF);
	}
					/* release io resources		*/
	if(sc->dsp_ioport) {
		if(ixjdebug > 0)
			device_printf(dev, "releasing ioport resources for DSP\n"); 
		bus_release_resource(dev, SYS_RES_IOPORT,
				sc->dsp_rid, sc->dsp_ioport);
	} else {
		device_printf(dev, "***No resources to release for DSP!\n"); 
	}

	destroy_dev(sc->dev);		/* remove device entry		*/
	
	return (0);
}

static device_method_t ixj_pci_methods[] = {
	/* device interface	*/
	DEVMETHOD(device_probe,		ixj_pci_probe),
	DEVMETHOD(device_attach,	ixj_pci_attach),
	DEVMETHOD(device_detach,	ixj_pci_detach),
	{ 0, 0 }
};

static driver_t ixj_pci_driver = {	/* the basic device interface	*/
	"ixj",
	ixj_pci_methods,
	sizeof(struct ixj_softc)
};

static	devclass_t	ixj_pci_devclass;

DRIVER_MODULE(ixj, pci, ixj_pci_driver, ixj_pci_devclass, 0, 0);
MODULE_VERSION(ixj, 1);

